# pylint: disable=cyclic-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.hardware.view.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Hardware Views."""

# Standard Library Imports
from copy import copy
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import (
    RAMSTK_ACTIVE_ENVIRONMENTS,
    RAMSTK_DORMANT_ENVIRONMENTS,
    RAMSTK_HR_DISTRIBUTIONS,
    RAMSTK_HR_MODELS,
    RAMSTK_HR_TYPES,
    RAMSTKUserConfiguration,
)
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.design_electric import (
    CapacitorDesignElectricInputPanel,
    ConnectionDesignElectricInputPanel,
    DesignElectricEnvironmentalInputPanel,
    DesignElectricStressInputPanel,
    DesignElectricStressResultPanel,
    ICDesignElectricInputPanel,
    InductorDesignElectricInputPanel,
    MeterDesignElectricInputPanel,
    MiscDesignElectricInputPanel,
    RelayDesignElectricInputPanel,
    ResistorDesignElectricInputPanel,
    SemiconductorDesignElectricInputPanel,
    SwitchDesignElectricInputPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    CapacitorMilHdbk217FResultPanel,
    ConnectionMilHdbk217FResultPanel,
    ICMilHdbk217FResultPanel,
    InductorMilHdbk217FResultPanel,
    MeterMilHdbk217FResultPanel,
    MiscellaneousMilHdbk217FResultPanel,
    RelayMilHdbk217FResultPanel,
    ResistorMilHdbk217FResultPanel,
    SemiconductorMilHdbk217FResultPanel,
    SwitchMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.reliability import (
    AvailabilityResultsPanel,
    ReliabilityInputPanel,
    ReliabilityResultsPanel,
)
from ramstk.views.gtk3.widgets import RAMSTKModuleView, RAMSTKPanel, RAMSTKWorkView

# RAMSTK Local Imports
from . import (
    HardwareGeneralDataPanel,
    HardwareLogisticsPanel,
    HardwareMiscellaneousPanel,
    HardwareTreePanel,
)


class HardwareModuleView(RAMSTKModuleView):
    """Display Hardware attribute data in the RAMSTK Module Book.

    The Hardware Module View displays all the Hardware associated with the
    connected RAMSTK Program in a flat list.  The attributes of a Hardware
    Module View are:

    :cvar _tag: the name of the module.

    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tablabel: str = "Hardware"
    _tabtooltip: str = _(
        "Displays the hardware hierarchy (BoM) for the selected Revision."
    )
    _tag: str = "hardware"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the Hardware Module View.

        :param configuration: the RAMSTK Configuration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.
        self._dic_icons["tab"] = (
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR + "/32x32/hardware.png"
        )

        # Initialize private list attributes.
        self._lst_callbacks[0] = self._do_request_insert_sibling
        self._lst_callbacks.insert(1, self._do_request_insert_child)
        self._lst_callbacks.insert(2, self._do_request_insert_part)
        self._lst_callbacks.insert(4, self._do_request_calculate_hardware)
        self._lst_callbacks.insert(5, self._do_request_calculate_all_hardware)
        self._lst_icons[0] = "insert_sibling"
        self._lst_icons.insert(1, "insert_child")
        self._lst_icons.insert(2, "insert_part")
        self._lst_icons.insert(4, "calculate")
        self._lst_icons.insert(5, "calculate_all")
        self._lst_mnu_labels: List[str] = [
            _("Add Sibling Assembly"),
            _("Add Child Assembly"),
            _("Add Piece Part"),
            _("Delete Selected"),
            _("Calculate the Selected"),
            _("Calculate the System"),
            _("Save Selected Hardware"),
            _("Save All Hardware"),
        ]
        self._lst_tooltips: List[str] = [
            _(
                "Adds a new hardware assembly at the same hierarchy level as the "
                "selected hardware item (i.e., a sibling hardware item)."
            ),
            _(
                "Adds a new hardware assembly one level subordinate to the selected "
                "hardware item (i.e., a child hardware item)."
            ),
            _(
                "Adds a new hardware component/piece-part to the the selected hardware "
                "assembly."
            ),
            _("Remove the currently selected hardware item and any children."),
            _("Calculate the selected hardware item and all of it's children."),
            _("Calculate the entire system."),
            _("Save changes to the selected hardware item."),
            _("Save changes to the entire system."),
        ]

        # Initialize private scalar attributes.
        self._pnlPanel = HardwareTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_set_record_id,
            f"selected_{self._tag}",
        )

    def _do_request_calculate_hardware(self, __button: Gtk.ToolButton) -> None:
        """Send request to calculate the selected hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_calculate_hardware",
            node_id=self.dic_pkeys["record_id"],
        )

    def _do_request_calculate_all_hardware(self, __button: Gtk.ToolButton) -> None:
        """Send request to iteratively calculate all hardware items.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_calculate_hardware",
            node_id=1,
        )

    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> Any:
        """Request to insert a new child assembly under the selected assembly.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _attributes = {
            "revision_id": self.dic_pkeys["revision_id"],
            "hardware_id": self.dic_pkeys["hardware_id"],
            "parent_id": self.dic_pkeys["hardware_id"],
            "part": 0,
            "record_id": self.dic_pkeys["record_id"],
        }

        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_insert_hardware",
            attributes=copy(_attributes),
        )
        _attributes.pop("part")
        pub.sendMessage(
            "request_insert_design_electric",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_design_mechanic",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_milhdbk217f",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_nswc",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_reliability",
            attributes=copy(_attributes),
        )

    def _do_request_insert_part(self, __button: Gtk.ToolButton) -> None:
        """Send request to insert a piece part to the selected Hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _attributes = {
            "revision_id": self.dic_pkeys["revision_id"],
            "hardware_id": self.dic_pkeys["hardware_id"],
            "parent_id": self.dic_pkeys["hardware_id"],
            "part": 1,
            "record_id": self.dic_pkeys["record_id"],
        }

        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_insert_hardware",
            attributes=copy(_attributes),
        )
        _attributes.pop("part")
        pub.sendMessage(
            "request_insert_design_electric",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_design_mechanic",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_milhdbk217f",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_nswc",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_reliability",
            attributes=copy(_attributes),
        )

    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> Any:
        """Send request to insert a new sibling Hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _attributes = {
            "revision_id": self.dic_pkeys["revision_id"],
            "hardware_id": self.dic_pkeys["hardware_id"],
            "parent_id": self.dic_pkeys["parent_id"],
            "part": 0,
            "record_id": self.dic_pkeys["record_id"],
        }

        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_insert_hardware",
            attributes=copy(_attributes),
        )
        _attributes.pop("part")
        pub.sendMessage(
            "request_insert_design_electric",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_design_mechanic",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_milhdbk217f",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_nswc",
            attributes=copy(_attributes),
        )
        pub.sendMessage(
            "request_insert_reliability",
            attributes=copy(_attributes),
        )

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the work stream module's record ID and, if any, parent ID.

        :param attributes: the attributes dict for the selected work stream module item.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["hardware_id"] = attributes["hardware_id"]
        self.dic_pkeys["parent_id"] = attributes["parent_id"]
        self.dic_pkeys["record_id"] = attributes["hardware_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the function module view.

        :return: None
        """
        super().make_ui()

        self._pnlPanel.do_set_cell_callbacks(
            "mvw_editing_hardware",
            [
                "alt_part_number",
                "cage_code",
                "cost",
                "description",
                "duty_cycle",
                "figure_number",
                "lcn",
                "manufacturer_id",
                "mission_time",
                "name",
                "nsn",
                "page_number",
                "part",
                "part_number",
                "quantity",
                "ref_des",
                "remarks",
                "repairable",
                "specification_number",
                "tagged_part",
                "year_of_manufacture",
            ],
        )
        self._pnlPanel.tvwTreeView.dic_handler_id[
            "button-press"
        ] = self._pnlPanel.tvwTreeView.connect(
            "button_press_event", super().on_button_press
        )
        for _element in ["assembly", "part"]:
            self._pnlPanel.dic_icons[_element] = self._dic_icons[_element]


class HardwareGeneralDataView(RAMSTKWorkView):
    """Display general Hardware attribute data in the RAMSTK Work Book.

    The Hardware Work View displays all the general data attributes for the
    selected Hardware.  The attributes of a Hardware General Data Work View
    are:

    :cvar _module: the name of the module.

    :ivar _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tag: str = "hardware"
    _tablabel: str = _("General\nData")
    _tabtooltip: str = _("Displays general information for the selected Hardware")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the Hardware Work View general data page.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :type configuration:
            :class:`ramstk.configuration.RAMSTKUserConfiguration`
        :param logger: the RAMSTKLogManager class instance.
        :type logger: :class:`ramstk.logger.RAMSTKLogManager`
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.
        self._dic_icons["comp_ref_des"] = (
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR + "/32x32/rollup.png"
        )

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self._pnlGeneralData: RAMSTKPanel = HardwareGeneralDataPanel()
        self._pnlLogistics: RAMSTKPanel = HardwareLogisticsPanel()
        self._pnlMiscellaneous: RAMSTKPanel = HardwareMiscellaneousPanel()

        self._lst_callbacks = [
            self._do_request_make_comp_ref_des,
            super().do_request_update,
            super().do_request_update_all,
        ]
        self._lst_icons = [
            "comp_ref_des",
            "save",
            "save-all",
        ]
        self._lst_mnu_labels = [
            _("Comp. Ref. Des."),
            _("Save"),
            _("Save All"),
        ]
        self._lst_tooltips = [
            _(
                "Creates the composite reference designator for the "
                "selected hardware item."
            ),
            _("Save changes to the currently selected hardware item."),
            _("Save changes to all hardware items."),
        ]

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_set_record_id,
            f"selected_{self._tag}",
        )

    def _do_request_make_comp_ref_des(self, __button: Gtk.ToolButton) -> None:
        """Send request to create the composite reference designator.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_make_comp_ref_des", node_id=self.dic_pkeys["record_id"]
        )
        super().do_set_cursor_active()

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the work stream module's record ID and, if any, parent ID.

        :param attributes: the attributes dict for the selected work stream
            module item.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["hardware_id"] = attributes["hardware_id"]
        self.dic_pkeys["parent_id"] = attributes["parent_id"]
        self.dic_pkeys["record_id"] = attributes["hardware_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the Hardware General Data tab.

        :return: None
        :rtype: None
        """
        _hpaned, _vpaned_right = super().do_make_layout_lrr()

        self._pnlGeneralData.fmt = self.fmt
        self._pnlGeneralData.dicSubcategories = (
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_SUBCATEGORIES
        )
        self._pnlGeneralData.do_load_categories(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_CATEGORIES
        )
        _hpaned.pack1(self._pnlGeneralData, True, True)

        self._pnlLogistics.do_load_cost_types()
        self._pnlLogistics.do_load_manufacturers(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_MANUFACTURERS
        )
        _vpaned_right.pack1(self._pnlLogistics, True, True)

        _vpaned_right.pack2(self._pnlMiscellaneous, True, True)

        self.show_all()


class HardwareAssessmentInputView(RAMSTKWorkView):
    """Display Hardware assessment input attribute data.

    The Hardware assessment input view displays all the assessment inputs for
    the selected Hardware item.  This includes, currently, inputs for
    MIL-HDBK-217FN2 and NSWC-11.  The attributes of a Hardware assessment
    input view are:

    :cvar list _lst_labels: the text to use for the assessment input widget
        labels.
    :cvar str _tag: the name of the module.

    :ivar dict _dic_assessment_input: dictionary of component-specific
        AssessmentInputs classes.
    :ivar int _hardware_id: the ID of the Hardware item currently being
        displayed.
    :ivar int _hazard_rate_method_id: the ID of the hazard rate method used for
        Hardware item.
    :ivar list _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar list _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar list _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar list _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dict attributes.

    # Define private list attributes.
    _lst_title: List[str] = [_("Operating Stresses")]

    # Define private scalar class attributes.
    _tag: str = "hardware"
    _tablabel: str = _("Assessment\nInputs")
    _tabtooltip: str = _(
        "Displays reliability assessment inputs for the selected hardware item."
    )

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Hardware assessment input view.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.
        self._dic_component_panels: Dict[int, RAMSTKPanel] = {
            1: ICDesignElectricInputPanel(),
            2: SemiconductorDesignElectricInputPanel(),
            3: ResistorDesignElectricInputPanel(),
            4: CapacitorDesignElectricInputPanel(),
            5: InductorDesignElectricInputPanel(),
            6: RelayDesignElectricInputPanel(),
            7: SwitchDesignElectricInputPanel(),
            8: ConnectionDesignElectricInputPanel(),
            9: MeterDesignElectricInputPanel(),
            10: MiscDesignElectricInputPanel(),
        }

        # Initialize private list attributes.
        self._lst_callbacks = [
            self._do_request_calculate,
            self._do_request_update,
            super().do_request_update_all,
        ]
        self._lst_icons = [
            "calculate",
            "save",
            "save-all",
        ]
        self._lst_tooltips = [
            _(
                "Calculate the currently selected hardware item and all of "
                "it's children."
            ),
            _("Save changes to the currently selected hardware item."),
            _("Save changes to all hardware items."),
        ]

        # Initialize private scalar attributes.
        self._pnlReliabilityInput: RAMSTKPanel = ReliabilityInputPanel()
        self._pnlEnvironmentalInput: RAMSTKPanel = (
            DesignElectricEnvironmentalInputPanel()
        )
        self._pnlStressInput: RAMSTKPanel = DesignElectricStressInputPanel()

        # We need to carry these as an attribute for this view because the
        # lower part of each is dynamically loaded with the component panels.
        self._vpnLeft: Gtk.VPaned = Gtk.VPaned()
        self._vpnRight: Gtk.VPaned = Gtk.VPaned()

        self._hazard_rate_method_id: int = 0
        self._subcategory_id: int = 0

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            super().do_set_cursor_active,
            "succeed_calculate_hardware",
        )

        pub.subscribe(
            self._do_pack_component_panel,
            "selected_hardware",
        )
        pub.subscribe(
            self._do_pack_component_panel,
            "succeed_get_hardware_attributes",
        )
        pub.subscribe(
            self._do_pack_component_panel,
            "hardware_category_changed",
        )
        pub.subscribe(
            self._do_set_record_id,
            f"selected_{self._tag}",
        )

    def _do_pack_component_panel(self, attributes: Dict[str, Any]) -> None:
        """Pack panel used to display component-specific input attributes.

        :param attributes: dict containing the attributes of the hardware
            item being loaded.
        :return: None
        :rtype: None
        """
        # If there was a component selected, hide it's widgets.  We get an
        # attribute error if no parts have been selected in the current
        # session.
        if self._vpnRight.get_child2() is not None:
            self._vpnRight.remove(self._vpnRight.get_child2())

        # Retrieve the appropriate component-specific view.
        if attributes["category_id"] > 0:
            _panel: RAMSTKPanel = self._dic_component_panels[attributes["category_id"]]
            _panel.fmt = self.fmt
            _panel.category_id = attributes["category_id"]
            self._vpnRight.pack2(_panel, True, True)
            self.show_all()

    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None:
        """Send request to calculate the selected hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        try:
            super().do_set_cursor_busy()
            pub.sendMessage(
                "request_calculate_hardware",
                node_id=self.dic_pkeys["record_id"],
            )
        except KeyError as _error:
            self.RAMSTK_LOGGER.do_log_exception(__name__, _error)

    def _do_request_update(self, __button: Gtk.ToolButton) -> None:
        """Send request to update all tables for the selected hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        for _table in [
            "design_electric",
            "hardware",
            "reliability",
        ]:
            pub.sendMessage(
                f"request_update_{_table}",
                node_id=self.dic_pkeys["record_id"],
            )

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the work stream module's record ID and, if any, parent ID.

        :param attributes: the attributes dict for the selected work stream
            module item.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["hardware_id"] = attributes["hardware_id"]
        self.dic_pkeys["parent_id"] = attributes["parent_id"]
        self.dic_pkeys["record_id"] = attributes["hardware_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the Hardware Assessment Input tab.

        :return: None
        :rtype: None
        """
        self._vpnLeft, self._vpnRight = super().do_make_layout_llrr()

        # Top left quadrant.
        self._pnlReliabilityInput.fmt = self.fmt
        self._pnlReliabilityInput.do_load_hr_distributions(RAMSTK_HR_DISTRIBUTIONS)
        self._pnlReliabilityInput.do_load_hr_methods(RAMSTK_HR_MODELS)
        self._pnlReliabilityInput.do_load_hr_types(RAMSTK_HR_TYPES)
        self._vpnLeft.pack1(self._pnlReliabilityInput, True, True)

        # Bottom left quadrant.
        self._pnlEnvironmentalInput.fmt = self.fmt
        self._pnlEnvironmentalInput.do_load_environment_active(
            RAMSTK_ACTIVE_ENVIRONMENTS
        )
        self._pnlEnvironmentalInput.do_load_environment_dormant(
            RAMSTK_DORMANT_ENVIRONMENTS
        )
        self._vpnLeft.pack2(self._pnlEnvironmentalInput, True, True)

        # Top right quadrant.
        self._pnlStressInput.fmt = self.fmt
        self._vpnRight.pack1(self._pnlStressInput, True, True)

        self.show_all()


class HardwareAssessmentResultsView(RAMSTKWorkView):
    """Display Hardware assessment results data in the RAMSTK Work View.

    The Hardware Assessment Results view displays all the assessment results
    for the selected Hardware.  The attributes of a Hardware Assessment Results
    View are:

    :cvar list _lst_labels: the text to use for the reliability assessment
        results widget labels.
    :cvar str _tag: the name of the module.

    :ivar dict _dic_assessment_results: dictionary of component-specific
        AssessmentResults classes.
    :ivar list _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar list _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar list _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar list _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private list class attributes.
    _lst_title = [_("Assessment Model Results"), _("Stress Results")]

    # Define private scalar class attributes.
    _tag: str = "hardware"
    _tablabel: str = _("Assessment\nResults")
    _tabtooltip: str = _(
        "Displays reliability, maintainability, and availability assessment results "
        "for the selected Hardware item."
    )

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Hardware assessment output view.

        :param configuration: the RAMSTK User Configuration class instance.
        :type configuration: :class:`ramstk.configuration.UserConfiguration`
        """
        super().__init__(configuration, logger)

        # Create a logger specifically for this class.
        self.RAMSTK_LOGGER.do_create_logger(
            __name__, self.RAMSTK_USER_CONFIGURATION.RAMSTK_LOGLEVEL, to_tty=False
        )

        # Initialize private dictionary attributes.
        self._dic_component_results: Dict[int, RAMSTKPanel] = {
            1: ICMilHdbk217FResultPanel(),
            2: SemiconductorMilHdbk217FResultPanel(),
            3: ResistorMilHdbk217FResultPanel(),
            4: CapacitorMilHdbk217FResultPanel(),
            5: InductorMilHdbk217FResultPanel(),
            6: RelayMilHdbk217FResultPanel(),
            7: SwitchMilHdbk217FResultPanel(),
            8: ConnectionMilHdbk217FResultPanel(),
            9: MeterMilHdbk217FResultPanel(),
            10: MiscellaneousMilHdbk217FResultPanel(),
        }

        # Initialize private list attributes.
        self._lst_callbacks = [
            self._do_request_calculate,
            super().do_request_update,
            super().do_request_update_all,
        ]
        self._lst_icons = [
            "calculate",
            "save",
            "save-all",
        ]
        self._lst_tooltips = [
            _("Calculate the currently selected Hardware item."),
            _("Save changes to the currently selected Hardware item."),
            _("Save changes to all Hardware items."),
        ]

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._subcategory_id: int = 0

        self._pnlAvailabilityResults: RAMSTKPanel = AvailabilityResultsPanel()
        self._pnlReliabilityResults: RAMSTKPanel = ReliabilityResultsPanel()
        self._pnlStressResults: RAMSTKPanel = DesignElectricStressResultPanel()

        # We need to carry these as an attribute for this view because the
        # lower part of each is dynamically loaded with the component panels.
        self._vpnLeft: Gtk.VPaned = Gtk.VPaned()
        self._vpnRight: Gtk.VPaned = Gtk.VPaned()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_pack_component_panel,
            "selected_hardware",
        )
        pub.subscribe(
            self._do_pack_component_panel,
            "succeed_get_hardware_attributes",
        )
        pub.subscribe(
            self._do_pack_component_panel,
            "hardware_category_changed",
        )
        pub.subscribe(
            self._do_set_record_id,
            f"selected_{self._tag}",
        )

    def _do_pack_component_panel(self, attributes: Dict[str, Any]) -> None:
        """Load the results specific to hardware components.

        :param attributes:
        :return: None
        :rtype: None
        """
        # If there was a component selected, hide it's widgets.  We get an
        # attribute error if no parts have been selected in the current
        # session.
        if self._vpnRight.get_child2() is not None:
            self._vpnRight.remove(self._vpnRight.get_child2())

        # Retrieve the appropriate component-specific view.
        if attributes["category_id"] > 0:
            _panel: RAMSTKPanel = self._dic_component_results[attributes["category_id"]]
            _panel.fmt = self.fmt
            self._vpnRight.pack2(_panel, True, True)
            self.show_all()

    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None:
        """Send request to calculate the selected Hardware item.

        :param __button: the Gtk.ToolButton() that called this method.
        :type __button: :class:Gtk.ToolButton`
        :return: None
        :rtype: None
        """
        try:
            super().do_set_cursor_busy()
            pub.sendMessage(
                "request_calculate_hardware",
                node_id=self.dic_pkeys["record_id"],
            )
        except (IndexError, KeyError) as _error:
            self.RAMSTK_LOGGER.do_log_exception(__name__, _error)
            _parent = self.get_parent().get_parent().get_parent().get_parent()
            _dialog = super().do_raise_dialog(parent=_parent)
            _dialog.do_set_message(
                _(
                    f"One or more inputs necessary to calculate "
                    f'hardware ID {self.dic_pkeys["hardware_id"]} is '
                    f"missing."
                )
            )
            _dialog.do_set_message_type("warning")
            _dialog.do_run()
            _dialog.do_destroy()
            pub.sendMessage(
                "fail_calculate_hardware",
                error_message=_error,
            )

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the work stream module's record ID and, if any, parent ID.

        :param attributes: the attribute dict for the selected work stream
            module item.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["hardware_id"] = attributes["hardware_id"]
        self.dic_pkeys["parent_id"] = attributes["parent_id"]
        self.dic_pkeys["record_id"] = attributes["hardware_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the Hardware Assessment Results tab.

        :return: False if successful or True if an error is encountered.
        :rtype: bool
        """
        self._vpnLeft, self._vpnRight = super().do_make_layout_llrr()

        # Top left quadrant.
        self._pnlReliabilityResults.fmt = self.fmt
        self._vpnLeft.pack1(self._pnlReliabilityResults, True, True)

        # Bottom left quadrant.
        self._pnlAvailabilityResults.fmt = self.fmt
        self._vpnLeft.pack2(self._pnlAvailabilityResults, True, True)

        # Top right quadrant.
        self._pnlStressResults.fmt = self.fmt
        self._vpnRight.pack1(self._pnlStressResults, True, True)

        self.show_all()



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.assistants.fmea.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2021 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK (D)FME(C)A Assistants Module."""

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKDialog, RAMSTKLabel


class AddStressTestMethod(RAMSTKDialog):
    """Assistant to walk user through process of adding stress or test."""

    def __init__(self, parent=None):
        """Initialize on instance of the Add Stress or Method Assistant."""
        super().__init__(
            _(
                "RAMSTK Physics of Failure Analysis Operating "
                "Stress and Test Method Addition Assistant"
            ),
            dlgparent=parent,
        )

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.rdoOpStress = Gtk.RadioButton.new_with_label_from_widget(
            None, _("Add stress")
        )
        self.rdoTestMethod = Gtk.RadioButton.new_from_widget(self.rdoOpStress)
        self.rdoTestMethod.set_label(_("Add test method"))

        self.__make_ui()

    def _cancel(self, __button: Gtk.Button) -> None:
        """Destroy the assistant when the 'Cancel' button is pressed.

        :param __button: the Gtk.Button() that called this method.
        """
        self.destroy()

    def __make_ui(self) -> None:
        """Build the user interface.

        :return: None
        :rtype: None
        """
        _fixed = Gtk.Fixed()
        self.vbox.pack_start(_fixed, True, True, 0)

        _label = RAMSTKLabel(
            _(
                "This is the RAMSTK Operating Stress and Test Method "
                "Addition Assistant.  Enter the information "
                "requested below and then press 'OK' to add "
                "a new design control or action to the RAMSTK "
                "Program database."
            )
        )
        _label.do_set_properties(width=600, height=-1, wrap=True)
        _fixed.put(_label, 5, 10)

        _y_pos: int = _label.get_preferred_size()[0].height + 50

        self.rdoOpStress.set_tooltip_text(
            _("Select to add an operating stress to the selected operating " "load.")
        )
        self.rdoTestMethod.set_tooltip_text(
            _("Select to add a test method to the selected operating load.")
        )

        _fixed.put(self.rdoOpStress, 10, _y_pos)
        _fixed.put(self.rdoTestMethod, 10, _y_pos + 35)

        _fixed.show_all()

        self.set_default_size(250, -1)



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.plot.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
# isort:skip_file
"""RAMSTK GTK3 Plot Module."""

# Standard Library Imports
from typing import Any, Dict, List, Tuple

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gdk, Gtk, _

# type: ignore
try:
    # noinspection PyPackageRequirements
    # Third Party Imports
    import matplotlib

    # noinspection PyPackageRequirements
    from matplotlib.backends.backend_gtk3cairo import (
        FigureCanvasGTK3Cairo as FigureCanvas,
    )

    # noinspection PyPackageRequirements
    from matplotlib.figure import Figure

    # noinspection PyPackageRequirements
    from matplotlib.lines import Line2D
except RuntimeError:
    # This is necessary to have the tests pass on headless servers.
    pass


class RAMSTKPlot:
    """The RAMSTKPlot class.

    This module contains RAMSTK plot class.  This class is derived from the
    applicable pyGTK widgets and matplotlib plots, but are provided with RAMSTK
    specific property values and methods.  This ensures a consistent look and
    feel to widgets in the RAMSTK application.
    """

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKPlot."""
        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_max: List[float] = []
        self._lst_min: List[float] = [0.0]

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.figure: matplotlib.figure.Figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.axis = self.figure.add_subplot(111)

    def do_load_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None:
        """Load the RAMSTKPlot.

        Plots can be one of the following types:

            * 'date'
            * 'histogram'
            * 'scatter' (default)
            * 'step'

        Pass the keyword plot_type to select.  Markers can be set using the
        marker keyword.  The default marker is 'g-' or a solid green line.  See
        matplotlib documentation for other options.

        :param x_values: list of the x-values to plot.
        :param y_values: list of the y-values to plot or list of bin edges if
            plotting a histogram.
        :return: None
        :rtype: None
        """
        _plot_type = kwargs.get("plot_type", "scatter")

        if _plot_type == "step":
            self._do_make_step_plot(x_values, y_values, **kwargs)
        elif _plot_type == "scatter":
            self._do_make_scatter_plot(x_values, y_values, **kwargs)
        elif _plot_type == "histogram":
            self._do_make_histogram(x_values, y_values, **kwargs)
        elif _plot_type == "date":
            self._do_make_date_plot(x_values, y_values, **kwargs)

        _min, _max = self._get_minimax_ordinates()

        self.axis.set_ybound(_min, 1.05 * _max)

        self.canvas.show()

    def do_add_line(
        self,
        x_values: List[float],
        y_values: List[float] = None,
        color: str = "k",
        marker: str = "^",
    ) -> None:
        """Load the RAMSTKPlot.

        :param x_values: list of the x-values to plot.
        :param y_values: list of the y-values to plot or list of bin edges if
            plotting a histogram.
        :param color: the color of the line to add to the plot.  Black is the
            default.  See matplotlib documentation for options.
        :param marker: the marker to use on the plot. Default is '^' or an
            upward pointing triangle.  See matplotlib documentation for other
            options.
        :return: None
        :rtype: None
        """
        _line = Line2D(
            x_values, y_values, lw=0.0, color=color, marker=marker, markersize=10
        )
        self.axis.add_line(_line)

    def do_close_plot(
        self, __window: Gtk.Window, __event: Gdk.Event, parent: Gtk.Widget
    ) -> None:
        """Return the plot to the Work Book page it is part of.

        :param __window: the Gtk.Window() that is being destroyed.
        :type __window: :class:`Gtk.Window`
        :param __event: the Gdk.Event() that called this method.
        :type __event: :class:`Gdk.Event`
        :param parent: the original parent Gtk.Widget() for the plot.
        :type parent: :class:`Gtk.Widget`
        :return: None
        :rtype: None
        """
        self.canvas.reparent(parent)

    # noinspection PyUnresolvedReferences
    def do_expand_plot(self, event: matplotlib.backend_bases.MouseEvent) -> None:
        """Display a plot in it's own window.

        :param event: the matplotlib.backend_bases.MouseEvent() that called
            this method.
        :type event: :class:`matplotlib.backend_bases.MouseEvent`
        :return: None
        :rtype: None
        """
        self.canvas = event.canvas
        _parent = self.canvas.get_parent()

        if event.button == 3:  # Right click.
            _window = Gtk.Window()
            _window.set_skip_pager_hint(True)
            _window.set_skip_taskbar_hint(True)
            _window.set_default_size(800, 400)
            _window.set_border_width(5)
            _window.set_position(Gtk.WindowPosition.NONE)
            _window.set_title(_("RAMSTK Plot"))

            _window.connect("delete_event", self.do_close_plot, _parent)

            self.canvas.reparent(_window)

            _window.show_all()

    # noinspection PyUnresolvedReferences
    def do_make_labels(self, label: str, **kwargs: Any) -> matplotlib.text.Text:
        """Make the abscissa or ordinate label.

        Accepts keyword arguments:
            * *fontsize* -- the size of the font to use for the axis label.
            * *fontweight* -- the weight of the font to use for the axis label.
            * *set_x* -- whether to set the abscissa (default) or ordinate
                label.
            * *x_pos* -- the position along the abscissa to place the label.
            * *y_pos* -- the position along the ordinate to place the label.

        :param label: the text to display as the abscissa or ordinate
            label.
        :return: matplotlib text instance representing the label.
        :rtype: :class:`matplotlib.text.Text`
        """
        _fontsize = kwargs.get("fontsize", 14)
        _fontweight = kwargs.get("fontweight", "bold")
        _set_x = kwargs.get("set_x", True)
        _x_pos = kwargs.get("x_pos", 0)
        _y_pos = kwargs.get("y_pos", 0)

        _label = None

        if _set_x:
            _label = self.axis.set_xlabel(
                label,
                {
                    "fontsize": _fontsize,
                    "fontweight": _fontweight,
                    "verticalalignment": "center",
                    "horizontalalignment": "center",
                    "x": _x_pos,
                    "y": _y_pos,
                },
            )
        else:
            _label = self.axis.set_ylabel(
                label,
                {
                    "fontsize": _fontsize,
                    "fontweight": _fontweight,
                    "verticalalignment": "center",
                    "horizontalalignment": "center",
                    "rotation": "vertical",
                },
            )

        return _label

    # pylint: disable=too-many-arguments
    def do_make_legend(self, text: List[str], **kwargs: Any) -> None:
        """Make a legend on the RAMSTKPlot.

        Accepts keyword arguments:
            * *fontsize* (str) -- the size of the font to use for the legend.
                Options are:
                    - xx-small
                    - x-small
                    - small (default)
                    - medium
                    - large
                    - x-large
                    - xx-large
            * *frameon* (bool) -- whether or not there is a frame around the
                legend.
            * *location* (str) -- the location of the legend on the plot.
                Options are:
                    - best
                    - upper right (default)
                    - upper left
                    - lower left
                    - lower right
                    - right
                    - center left
                    - center right
                    - lower center
                    - upper center
                    - center
            * *ncol* (int) -- the number columns in the legend.  Default is 1.
            * *shadow* (bool) -- whether or not to display a shadow behind the
                legend block.  Default is True.
            * *title* (str) -- the title of the legend.  Default is an empty
                string.
            * *lwd* (float) -- the linewidth of the box around the legend.

        :param text: the text to display in the legend.
        :return: None
        :rtype: None
        """
        _fontsize = kwargs.get("fontsize", "small")
        _frameon = kwargs.get("frameon", False)
        _location = kwargs.get("location", "upper right")
        _lwd = kwargs.get("lwd", 0.5)
        _ncol = kwargs.get("ncol", 1)
        _shadow = kwargs.get("shadow", True)
        _title = kwargs.get("title", "")

        _legend = self.axis.legend(
            text,
            frameon=_frameon,
            loc=_location,
            ncol=_ncol,
            shadow=_shadow,
            title=_title,
        )

        for _text in _legend.get_texts():
            _text.set_fontsize(_fontsize)
        for _line in _legend.get_lines():
            _line.set_linewidth(_lwd)

    # noinspection PyUnresolvedReferences
    def do_make_title(
        self, title: str, fontsize: int = 16, fontweight: str = "bold"
    ) -> matplotlib.text.Text:
        """Make the plot title.

        :param title: the text to display as the title.
        :param fontsize: the size of the font to use for the title.
        :param fontweight: the weight of the font to use for the title.
        :return: matplotlib text instance representing the title.
        :rtype: :class:`matplotlib.text.Text`
        """
        return self.axis.set_title(
            title,
            {
                "fontsize": fontsize,
                "fontweight": fontweight,
                "verticalalignment": "baseline",
                "horizontalalignment": "center",
            },
        )

    def _do_make_date_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None:
        """Make a date plot.

        Use the keyword marker to set the type and color of marker to use
        for the plot.  Default is 'g-' or a solid green line.

        :param x_values: the list of x-values (dates) for the plot.
        :param y_values: the list of y-values for the plot.
        :return: None
        :rtype: None
        """
        _marker = kwargs.get("marker", "g-")

        if y_values is not None:
            self.axis.plot_date(x_values, y_values, _marker, xdate=True, linewidth=2)
            self._lst_min.append(min(y_values))
            self._lst_max.append(max(y_values))

    def _do_make_histogram(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None:
        """Make a histogram.

        Use the keyword marker to set the color of the bars used for the plot.
        Default is 'g' or green.

        :param list x_values: the list of x-values for the plot.
        :param list y_values: the list of bin edges for the plot.
        :keyword str marker: the color of bars to use for the histogram.
            Default is green.
        :return: None
        :rtype: None
        """
        _marker = kwargs.get("marker", "g")

        if y_values is not None:
            self.axis.grid(False, which="both")
            # pylint: disable=unused-variable
            _values, _edges, __ = self.axis.hist(x_values, bins=y_values, color=_marker)
            self._lst_min.append(min(_values))
            self._lst_max.append(max(_values) + 1)

    def _do_make_scatter_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None:
        """Make a scatter plot.

        Use the keyword marker to set the type and color of marker to use
        for the plot.  Default is 'go' or open green circles.

        :param x_values: the list of x-values for the plot.
        :param y_values: the list of y-values for the plot.
        :return: None
        :rtype: None
        """
        _marker = kwargs.get("marker", "go")

        if y_values is not None:
            (_line,) = self.axis.plot(x_values, y_values, _marker, linewidth=2)
            _line.set_ydata(y_values)
            self._lst_min.append(min(y_values))
            self._lst_max.append(max(y_values))

    def _do_make_step_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None:
        """Make a step plot.

        Use the keyword marker to set the type and color of marker to use
        for the plot.  Default is 'g-' or a solid green line.

        :param list x_values: the list of x-values for the plot.
        :param list y_values: the list of y-values for the plot.
        :return: None
        :rtype: None
        """
        _marker = kwargs.get("marker", "g-")

        if y_values is not None:
            (_line,) = self.axis.step(x_values, y_values, _marker, where="mid")
            _line.set_ydata(y_values)
            self._lst_min.append(min(y_values))
            self._lst_max.append(max(y_values))

    def _get_minimax_ordinates(self) -> Tuple[float, float]:
        """Get minimum and maximum y-values to set the axis bounds.

        If the maximum value is infinity, use the next largest value and so
        forth.

        :return: _min, _max; tuple containing the minimum and maximum ordinate
            values.
        """
        _min: float = min(self._lst_min)
        _max: float = max(1.0, self._lst_max[0])
        for i in range(1, len(self._lst_max)):
            if _max < self._lst_max[i] != float("inf"):
                _max = self._lst_max[i]

        return _min, _max



# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord

class RAMSTKMissionPhaseRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Dict[str, Any]
    __tablename__: str
    __table_args__: Dict[str, bool]
    revision_id: int
    mission_id: int
    phase_id: int
    description: str
    name: str
    phase_start: float
    phase_end: float
    mission: relationship
    environment: relationship
    is_mission: bool
    is_phase: bool
    is_env: bool
    def get_attributes(self) -> Dict[str, Any]: ...



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.assistants.__init__.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Books Package."""

# RAMSTK Local Imports
from .export import ExportProject
from .fmea import AddControlAction
from .imports import ImportProject
from .pof import AddStressTestMethod
from .project import CreateProject, OpenProject
from .similaritem import EditFunction



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.site_info.table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Site Information Table Model."""

# Standard Library Imports
from typing import Type

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseTable, RAMSTKSiteInfoRecord


class RAMSTKSiteInfoTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Option data manager."""

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_site_id"
    _db_tablename = "ramstk_site_info"
    _select_msg = "request_get_option_attributes2"
    _tag = "option"

    # Define public dict class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs) -> None:
        """Initialize a Options data manager instance."""
        RAMSTKBaseTable.__init__(self, **kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "site_id",
        ]

        # Initialize private scalar attributes.
        self._record: Type[RAMSTKSiteInfoRecord] = RAMSTKSiteInfoRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "site_id"

        # Subscribe to PyPubSub messages.



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.milhdbk217f.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""milhdbk217f Calculations Class."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Local Imports
from .models import (
    capacitor,
    connection,
    crystal,
    efilter,
    fuse,
    inductor,
    integratedcircuit,
    lamp,
    meter,
    relay,
    resistor,
    semiconductor,
    switch,
)


# noinspection PyTypeChecker
def do_predict_active_hazard_rate(**attributes: Dict[str, Any]) -> float:
    """Calculate the active hazard rate for a hardware item.

    .. attention:: The programmer is responsible for ensuring appropriate
        stress analyses (e.g., voltage ratios) are performed and results
        assigned to the attribute dict prior to calling the MIL-HDBK-217F
        methods.

    .. important:: The calling object is responsible for handling any
        exceptions raised by or passed through this method.

    :return: attributes['hazard_rate_active']
    :rtype: float
    """
    try:
        if attributes["hazard_rate_method_id"] == 1:
            attributes = _do_calculate_part_count(**attributes)
        elif attributes["hazard_rate_method_id"] == 2:
            attributes = _do_calculate_part_stress(**attributes)

        pub.sendMessage("succeed_predict_reliability", attributes=attributes)
        pub.sendMessage("request_set_all_milhdbk217f_attributes", attributes=attributes)
        pub.sendMessage("request_set_all_reliability_attributes", attributes=attributes)
    except ValueError:
        pub.sendMessage(
            "fail_predict_reliability",
            error_message=(
                f"Failed to predict MIL-HDBK-217F hazard rate for hardware ID "
                f'{attributes["hardware_id"]}; one or more inputs has a negative or '
                f"missing value. Hardware item category "
                f'ID={attributes["category_id"]}, subcategory '
                f'ID={attributes["subcategory_id"]}, rated '
                f'power={attributes["power_rated"]}, number of '
                f'elements={attributes["n_elements"]}.'
            ),
        )
    except ZeroDivisionError:
        pub.sendMessage(
            "fail_predict_reliability",
            error_message=(
                f"Failed to predict MIL-HDBK-217F hazard rate for hardware ID "
                f'{attributes["hardware_id"]}; one or more inputs has a value of 0.0.  '
                f'Hardware item category ID={attributes["category_id"]}, subcategory '
                f'ID={attributes["subcategory_id"]}, operating ac '
                f'voltage={attributes["voltage_ac_operating"]}, operating DC '
                f'voltage={attributes["voltage_dc_operating"]}, operating '
                f'temperature={attributes["temperature_active"]}, temperature '
                f'rise={attributes["temperature_rise"]}, rated maximum '
                f'temperature={attributes["temperature_rated_max"]}, feature '
                f'size={attributes["feature_size"]}, surface '
                f'area={attributes["area"]}, and item weight={attributes["weight"]}.'
            ),
        )

    return attributes["hazard_rate_active"]


# noinspection PyTypeChecker
def _do_calculate_part_count(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the MIL-HDBK-217F parts count active hazard rate.

    :param attributes: the attributes dict for the component being
        calculated.
    :return: attributes; the attributes dict with updated values.
    :rtype: dict
    :raise: IndexError if there is no entry for the active environment ID.
    :raise: KeyError if there is no entry for category ID or subcategory ID.
    """
    _part_count = {
        1: integratedcircuit.calculate_part_count,
        2: semiconductor.calculate_part_count,
        3: resistor.calculate_part_count,
        4: capacitor.calculate_part_count,
        5: inductor.calculate_part_count,
        6: relay.calculate_part_count,
        7: switch.calculate_part_count,
        8: connection.calculate_part_count,
        9: meter.calculate_part_count,
        10: {
            1: crystal.calculate_part_count,
            2: efilter.calculate_part_count,
            3: fuse.calculate_part_count,
            4: lamp.calculate_part_count,
        },
    }

    if attributes["category_id"] == 2:
        attributes = _part_count[attributes["category_id"]](**attributes)
    elif attributes["category_id"] == 10:
        attributes["lambda_b"] = _part_count[attributes["category_id"]][
            attributes["subcategory_id"]
        ](**attributes)
    else:
        attributes["lambda_b"] = _part_count[attributes["category_id"]](**attributes)

    if attributes["category_id"] != 2:
        attributes["piQ"] = _get_part_count_quality_factor(
            attributes["category_id"],
            attributes["subcategory_id"],
            attributes["quality_id"],
        )
    else:
        attributes = semiconductor.get_part_count_quality_factor(attributes)

    attributes["hazard_rate_active"] = attributes["lambda_b"] * attributes["piQ"]

    return attributes


# noinspection PyTypeChecker
def _do_calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the MIL-HDBK-217F parts stress active hazard rate.

    :param attributes: the attributes dict for the component being
        calculated.
    :return: attributes; the attributes dict with updated values.
    :rtype: dict
    :raise: IndexError if there is no entry for the active environment ID.
    :raise: KeyError if there is no entry for category ID or subcategory ID.
    """
    _functions = {
        1: integratedcircuit.calculate_part_stress,
        2: semiconductor.calculate_part_stress,
        3: resistor.calculate_part_stress,
        4: capacitor.calculate_part_stress,
        5: inductor.calculate_part_stress,
        6: relay.calculate_part_stress,
        7: switch.calculate_part_stress,
        8: connection.calculate_part_stress,
        9: meter.calculate_part_stress,
        10: {
            1: crystal.calculate_part_stress,
            2: efilter.calculate_part_stress,
            3: fuse.calculate_part_stress,
            4: lamp.calculate_part_stress,
        },
    }

    if attributes["category_id"] != 6:
        attributes["piE"] = _get_environment_factor(
            attributes["category_id"],
            attributes["environment_active_id"],
            subcategory_id=attributes["subcategory_id"],
            quality_id=attributes["quality_id"],
        )

    if attributes["category_id"] not in [2, 5]:
        attributes["piQ"] = _get_part_stress_quality_factor(
            attributes["category_id"],
            attributes["subcategory_id"],
            attributes["quality_id"],
        )

    if attributes["category_id"] == 10:
        _part_stress = _functions[attributes["category_id"]][
            attributes["subcategory_id"]
        ]
    else:
        _part_stress = _functions[attributes["category_id"]]

    return _part_stress(**attributes)


def _get_environment_factor(
    category_id: int,
    environment_active_id: int,
    subcategory_id: int = -1,
    quality_id: int = -1,
) -> float:
    """Retrieve the MIL-HDBK-217F environment factor (piE) for the component.

    Most component types have a single list of piE factors, but some require
    additional indices to select the correct list of factors.

    :param category_id: the category ID of the component.
    :param environment_active_id: the active environment ID for the
        component.
    :keyword int subcategory_id: the subcategory ID of the component.
    :keyword int quality_id: the quality level ID of the component.
    :return: _pi_e; the selected piE value.
    :rtype: float
    :raise: IndexError if there is no list entry for the passed active
        environment ID.
    :raise: KeyError if there is no piE list for the passed category ID (or
        subcategory ID, quality ID when appllicable).
    """
    _pi_e_lists = {
        1: integratedcircuit.PI_E,
        2: semiconductor.PI_E,
        3: resistor.PI_E,
        4: capacitor.PI_E,
        5: inductor.PI_E,
        7: switch.PI_E,
        8: connection.PI_E,
        9: meter.PI_E,
        10: {1: crystal.PI_E, 2: efilter.PI_E, 3: fuse.PI_E, 4: lamp.PI_E},
    }

    if category_id == 8 and subcategory_id in [1, 2]:
        _pi_e = _pi_e_lists[category_id][subcategory_id][quality_id][
            environment_active_id - 1
        ]
    elif category_id in [2, 3, 5, 7, 9, 10] or (
        category_id == 8 and subcategory_id not in [1, 2]
    ):
        _pi_e = _pi_e_lists[category_id][subcategory_id][environment_active_id - 1]
    else:
        _pi_e = _pi_e_lists[category_id][environment_active_id - 1]

    return _pi_e


def _get_part_count_quality_factor(
    category_id: int, subcategory_id: int, quality_id: int
) -> float:
    """Retrieve the MIL-HDBK-217F parts count quality factor (piQ).

    .. note:: Fuses and Lamps have no piQ input.

    .. note:: Semiconductors have a more complicated piQ listing and the
        function to select the correct value is included in the semiconductor
        model file.

    :param category_id: the category ID of the component.
    :param subcategory_id: the subcategory ID of the component.
    :param quality_id: the quality level ID for the component.
    :return: _pi_q; the selected piQ value.
    :rtype: float
    :raise: IndexError if there is no list entry for the passed active
        environment ID.
    :raise: KeyError if there is no piQ list for the passed category ID.
    """
    _pi_q_lists = {
        1: integratedcircuit.PI_Q,
        3: resistor.PART_COUNT_PI_Q,
        4: capacitor.PART_COUNT_PI_Q,
        5: inductor.PART_COUNT_PI_Q,
        6: relay.PART_COUNT_PI_Q,
        7: switch.PART_COUNT_PI_Q,
        8: connection.PART_COUNT_PI_Q,
        9: meter.PART_COUNT_PI_Q,
        10: {1: crystal.PART_COUNT_PI_Q, 2: efilter.PI_Q},
    }

    if category_id in [6, 7, 9]:
        _pi_q = _pi_q_lists[category_id][subcategory_id][quality_id - 1]
    elif category_id == 10 and subcategory_id in [1, 2]:
        _pi_q = _pi_q_lists[category_id][subcategory_id][quality_id - 1]
    elif category_id == 10 and subcategory_id in [3, 4]:
        _pi_q = 1.0
    else:
        _pi_q = _pi_q_lists[category_id][quality_id - 1]

    return _pi_q


def _get_part_stress_quality_factor(
    category_id: int, subcategory_id: int, quality_id: int
) -> float:
    """Retrieve the MIL-HDBK-217F part stress quality factor (piQ).

    .. note:: Fuses and Lamps have no piQ input.

    :param category_id: the category ID of the component.
    :param subcategory_id: the subcategory ID of the component.
    :param quality_id: the quality level ID for the component.
    :return: _pi_q; the selected piQ value.
    :rtype: float
    :raise: IndexError if there is no list entry for the passed quality ID.
    :raise: KeyError if there is no piQ list for the passed category ID.
    """
    _pi_q_lists = {
        1: integratedcircuit.PI_Q,
        3: resistor.PART_STRESS_PI_Q,
        4: capacitor.PART_STRESS_PI_Q,
        6: relay.PART_STRESS_PI_Q,
        7: switch.PART_STRESS_PI_Q,
        8: connection.PART_STRESS_PI_Q,
        9: meter.PART_STRESS_PI_Q,
        10: {1: crystal.PART_STRESS_PI_Q, 2: efilter.PI_Q},
    }

    if category_id == 1:
        _pi_q = _pi_q_lists[category_id][quality_id - 1]
    elif (category_id == 8 and subcategory_id in [4, 5]) or (
        category_id == 7 and subcategory_id == 5
    ):
        _pi_q = _pi_q_lists[category_id][subcategory_id][quality_id - 1]
    elif category_id == 7 and subcategory_id != 5:
        _pi_q = 0.0
    elif category_id == 8 and subcategory_id not in [4, 5]:
        _pi_q = 0.0
    elif category_id == 9 and subcategory_id == 1:
        _pi_q = 0.0
    elif category_id == 10 and subcategory_id in [3, 4]:
        _pi_q = 0.0
    else:
        _pi_q = _pi_q_lists[category_id][subcategory_id][quality_id - 1]

    return _pi_q



# pylint: disable=non-parent-init-called, too-many-public-methods, cyclic-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK GTK3 Panel Class."""

# Standard Library Imports
import inspect
from typing import Any, Dict, List, Union

# Third Party Imports
# pylint: disable=ungrouped-imports
# noinspection PyPackageValidations
import treelib
from pandas.plotting import register_matplotlib_converters
from pubsub import pub

# RAMSTK Package Imports
from ramstk.utilities import boolean_to_integer
from ramstk.views.gtk3 import Gtk, _

# RAMSTK Local Imports
from .button import RAMSTKCheckButton
from .combo import RAMSTKComboBox
from .entry import RAMSTKEntry, RAMSTKTextView
from .frame import RAMSTKFrame
from .label import RAMSTKLabel, do_make_label_group
from .plot import RAMSTKPlot
from .scrolledwindow import RAMSTKScrolledWindow
from .treeview import RAMSTKTreeView

register_matplotlib_converters()


class RAMSTKPanel(RAMSTKFrame):
    """The RAMSTKPanel class.

    The attributes of a RAMSTKPanel are:

    :cvar _record_field: the database table field that contains the record ID.
    :cvar _select_msg: the PyPubSub message the panel listens for to load values into
        it's attribute widgets.  Defaults to "selected_revision".
    :cvar _tag: the name of the tag in the table or view model tree.  This should be
        the same value as the _tag attribute for the table or view model the panel is
        used to display.
    :cvar _title: the title to display on the panel frame.

    :ivar _parent_id: the ID of the parent entity for the selected work stream entity.
        This is needed for hierarchical modules such as the function module.  For flat
        modules, this will always be zero.
    :ivar _record_id: the work stream module ID whose attributes this panel is
        displaying.

    :ivar dic_attribute_widget_map: a dict used to map model attributes to their
        respective display widgets.  The key is the name of the attribute in the record
        model and the value is a list containing the following information:

            * Position 0 is the (zero-based) index of the widget.
            * Position 1 is the widget used to display the attribute.  This can be a
                named instance of a widget (e.g., self.txtName) or a widget class from
                which an unnamed instance will be created
                (e.g., Gtk.CellRendererText()).
            * Position 2 is the signal emitted by the widget when it is updated/edited.
            * Position 3 is the callback method that emits the signal in position 2.
                Set this to None to make widget read-only.
            * Position 4 is the PyPubSub message published when the widget is updated.
            * Position 5 is the default value to display in the widget.
            * Position 6 is a dict containing the property values for the widget.
            * Position 7 is the text to use for the widget in position 1 label.  For
                RAMSTKTreeViews this will be the column heading.
            * Position 8 is the Gobject data type to display in the widget.

        For a fixed panel, dict entries should be in the order they should appear in
        the panel.

    :ivar fmt: the formatting string for displaying float values.
    """

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field: str = ""
    _select_msg: str = "selected_revision"
    _tag: str = ""
    _title: str = ""

    # Define public dict class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKPanel.

        :return: None
        :rtype: None
        """
        super().__init__()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._parent_id: int = -1
        self._record_id: int = -1

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {}

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.
        self.fmt: str = "{0:0.6}"

        # Subscribe to PyPubSub messages.


class RAMSTKFixedPanel(RAMSTKPanel):
    """The RAMSTKFixedPanel class."""

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.

    # Define public dict class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKFixedPanel.

        :return: None
        :rtype: None
        """
        super().__init__()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_clear_panel,
            "request_clear_views",
        )
        pub.subscribe(
            self.do_load_panel,
            self._select_msg,
        )
        pub.subscribe(self.on_edit, f"mvw_editing_{self._tag}")

        # Generally used with panels that accept inputs and are, thus, editable.
        try:
            pub.subscribe(
                self._do_set_sensitive,
                f"succeed_get_{self._tag}_attributes",
            )
        except AttributeError:
            pass

        # Generally used with panels that display results and are, thus, uneditable.
        try:
            pub.subscribe(
                self._do_load_entries,
                f"succeed_get_{self._tag}_attributes",
            )
        except AttributeError:
            pass

    def do_clear_panel(self) -> None:
        """Clear the contents of the widgets on a fixed type panel.

        :return: None
        :rtype: None
        """
        for (
            __,  # pylint: disable=unused-variable
            _value,
        ) in self.dic_attribute_widget_map.items():
            _value[1].do_update(_value[4], signal=_value[2])

    def do_load_panel(
        self,
        attributes: Dict[str, Any],
    ) -> None:
        """Load data into the widgets on fixed type panel.

        :param attributes: the attributes dict for the selected item.
        :return: None
        """
        self._record_id = attributes[self._record_field]

        for _key, _value in self.dic_attribute_widget_map.items():
            _value[1].do_update(
                attributes.get(_key, _value[5]),
                signal=_value[2],
            )

        pub.sendMessage("request_set_cursor_active")

    def do_make_panel(self, **kwargs: Dict[str, Any]) -> None:
        """Create a panel with the labels and widgets on a Gtk.Fixed().

        :return: None
        :rtype: None
        """
        _justify = kwargs.get("justify", Gtk.Justification.RIGHT)

        # Extract the list of labels and associated widgets from the attribute-widget
        # map.
        _lst_labels = [x[1][7] for x in self.dic_attribute_widget_map.items()]
        _lst_widgets = [x[1][1] for x in self.dic_attribute_widget_map.items()]

        _fixed: Gtk.Fixed = Gtk.Fixed()

        _y_pos: int = 5
        # noinspection PyTypeChecker
        (_x_pos, _labels) = do_make_label_group(
            _lst_labels,
            bold=False,  # type: ignore
            justify=_justify,
            x_pos=5,  # type: ignore
            y_pos=5,  # type: ignore
        )
        for _idx, _label in enumerate(_labels):
            _fixed.put(_label, 5, _y_pos)

            _minimum: Gtk.Requisition = _lst_widgets[  # type: ignore
                _idx
            ].get_preferred_size()[0]
            if _minimum.height <= 0:
                _minimum.height = _lst_widgets[_idx].height  # type: ignore

            # RAMSTKTextViews are placed inside a scrollwindow so that's
            # what needs to be placed on the container.
            if isinstance(_lst_widgets[_idx], RAMSTKTextView):
                _fixed.put(
                    _lst_widgets[_idx].scrollwindow,  # type: ignore
                    _x_pos + 10,
                    _y_pos,
                )
                _y_pos += _minimum.height + 30
            elif isinstance(_lst_widgets[_idx], RAMSTKCheckButton):
                _fixed.put(_lst_widgets[_idx], _x_pos + 10, _y_pos)
                _y_pos += _minimum.height + 30
            else:
                _fixed.put(_lst_widgets[_idx], _x_pos + 10, _y_pos)
                _y_pos += _minimum.height + 5

        _scrollwindow: RAMSTKScrolledWindow = RAMSTKScrolledWindow(_fixed)

        self.add(_scrollwindow)

    def do_set_callbacks(self) -> None:
        """Set the callback methods for RAMSTKTreeView().

        :return: None
        """
        for (
            _key,
            _value,
        ) in self.dic_attribute_widget_map.items():
            # Value[1] is the widget who's callback is being set.
            _value[1].dic_handler_id[_value[2]] = _value[1].connect(
                _value[2],  # Widget's callback signal.
                _value[3],  # Callback function/method.
                _key,  # Attribute name.
                _value[4],  # Message to for callback to emit on success.
            )

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set properties of the RAMSTKPanel() widgets.

        :return: None
        """
        super().do_set_properties(**{"bold": True, "title": self._title})

        for (
            __,  # pylint: disable=unused-variable
            _value,
        ) in self.dic_attribute_widget_map.items():
            _value[1].do_set_properties(**_value[6])

    def on_changed_combo(
        self, combo: RAMSTKComboBox, key: str, message: str
    ) -> Dict[Union[str, Any], Any]:
        """Retrieve changes made in RAMSTKComboBox() widgets.

        This method publishes the PyPubSub message that it is passed.  This
        is usually sufficient to ensure the attributes are updated by the
        datamanager.  This method also return a dict with {_key: _new_text}
        if this information is needed by the child class.

        :param combo: the RAMSTKComboBox() that called the method.
        :param key: the name in the class' Gtk.TreeModel() associated
            with the attribute from the calling Gtk.Widget().
        :param message: the PyPubSub message to publish.
        :return: {_key: _new_text}; the work stream module's attribute name
            and the new value from the RAMSTKComboBox().  The value {'': -1}
            will be returned when a KeyError or ValueError is raised by this
            method.
        """
        _key: str = ""
        _new_text: int = -1

        combo.handler_block(combo.dic_handler_id["changed"])

        try:
            _new_text = int(combo.get_active())

            # Only if something is selected should we send the message.
            # Otherwise, attributes get updated to a value of -1 which isn't
            # correct.  And it sucks trying to figure out why, so leave the
            # conditional unless you have a more elegant (and there prolly
            # is) solution.
            if _new_text > -1:
                pub.sendMessage(
                    message,
                    node_id=self._record_id,
                    package={key: _new_text},
                )
        except (KeyError, ValueError):
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while editing {self._tag} data "
                f"for record ID {self._record_id} in the view.  Key {key} does not "
                f"exist in attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

        combo.handler_unblock(combo.dic_handler_id["changed"])

        return {_key: _new_text}

    def on_changed_entry(
        self, entry: RAMSTKEntry, key: str, message: str
    ) -> Dict[Union[str, Any], Any]:
        """Retrieve changes made in RAMSTKEntry() widgets.

        This method is called by:

            * RAMSTKEntry() 'changed' signal

        This method publishes the PyPubSub message that it is passed.  This
        is usually sufficient to ensure the attributes are updated by the
        datamanager.  This method also return a dict with {_key: _new_text}
        if this information is needed by the child class.

        :param entry: the RAMSTKEntry() that called the method.
        :param key: the name in the class' Gtk.TreeModel() associated
            with the data from the calling RAMSTKEntry() or RAMSTKTextView().
        :param message: the PyPubSub message to publish.
        :return: {_key: _new_text}; the child module attribute name and the
            new value from the RAMSTKEntry() or RAMSTKTextView(). The value
            {'': ''} will be returned when a KeyError or ValueError is raised
            by this method.
        :rtype: dict
        """
        try:
            _handler_id = entry.dic_handler_id["changed"]
        except KeyError:
            _handler_id = entry.dic_handler_id["value-changed"]

        entry.handler_block(_handler_id)

        _package: Dict[str, Any] = self.__do_read_text(
            entry, key, self.dic_attribute_widget_map[key][8]
        )

        entry.handler_unblock(_handler_id)

        pub.sendMessage(message, node_id=self._record_id, package=_package)

        return _package

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def on_changed_textview(
        self, buffer: Gtk.TextBuffer, key: str, message: str, textview: RAMSTKTextView
    ) -> Dict[Union[str, Any], Any]:
        """Retrieve changes made in RAMSTKTextView() widgets.

        This method is called by:

            * Gtk.TextBuffer() 'changed' signal

        :param buffer: the Gtk.TextBuffer() calling this method.  This
            parameter is unused in this method.
        :param key: the name in the class' Gtk.TreeModel() associated
            with the data from the calling RAMSTKTextView().
        :param message: the PyPubSub message to broadcast.
        :param textview: the RAMSTKTextView() calling this method.
        :return: {_key: _new_text}; the child module attribute name and the
            new value from the RAMSTKTextView(). The value {'': ''} will be
            returned when a KeyError or ValueError is raised by this method.
        """
        textview.handler_block(textview.dic_handler_id["changed"])

        _package: Dict[str, Any] = self.__do_read_text(
            textview, key, self.dic_attribute_widget_map[key][8]
        )

        textview.handler_unblock(textview.dic_handler_id["changed"])

        pub.sendMessage(message, node_id=[self._record_id, -1], package=_package)

        return _package

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def on_edit(self, node_id: List[int], package: Dict[str, Any]) -> None:
        """Update the panel's Gtk.Widgets() when attributes are changed.

        This method is used to update a RAMSTKPanel() containing a
        Gtk.Fixed() populated with widgets [generally the work view]
        whenever a module view RAMSTKTreeView() is edited.  It is used to keep
        the data displayed in-sync.

        A dict 'package' is sent when a module view RAMSTKTreeView() is
        edited/changed.

            `package` key: `package` value

        corresponds to:

            database field name: database field new value

        The key in the 'package' is used to find the value in
        _dic_attribute_updater corresponding to the data being changed.
        Position 0 of the _dic_attribute_updater value list is the method
        used to update the widget and position 1 is the name of the signal
        to block during the update.

        :param node_id: the list of IDs of the work stream module item
            being edited.  This unused parameter is part of the PyPubSub
            message data package that this method responds to so it must
            remain in the argument list.
        :param package: a dict containing the attribute name as key and
            the new attribute value as the value.
        :return: None
        """
        _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
        [[_key, _value]] = package.items()

        try:
            _widget = self.dic_attribute_widget_map[_key][1]
            _signal = self.dic_attribute_widget_map[_key][2]
            _widget.do_update(_value, _signal)  # type: ignore
        except KeyError:
            _error_msg = _(
                f"{_method_name}: An error occurred while updating {self._tag} data "
                f"for record ID {self._record_id} in the view.  No key {_key} in "
                f"dic_attribute_widget_map."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
        except TypeError:
            _error_msg = _(
                f"{_method_name}: An error occurred while updating {self._tag} data "
                f"for record ID {self._record_id} in the view.  Data for key {_key} is "
                f"the wrong type."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

    def on_toggled(
        self, checkbutton: RAMSTKCheckButton, key: str, message: str
    ) -> Dict[Union[str, Any], Any]:
        """Retrieve changes made in RAMSTKCheckButton() widgets.

        :param checkbutton: the RAMSTKCheckButton() that was toggled.
        :param key: the name in the class' Gtk.TreeModel() associated
            with the data from the calling RAMSTKCheckButton().
        :param message: the PyPubSub message to broadcast.
        :return: {_key: _new_text}; the child module attribute name and the
            new value from the RAMSTKEntry() or RAMSTKTextView(). The value
            {'': -1} will be returned when a KeyError is raised by this method.
        """
        _new_text: int = -1

        try:
            _new_text = int(checkbutton.get_active())
            checkbutton.do_update(_new_text, signal="toggled")

            pub.sendMessage(
                message,
                node_id=[self._record_id, -1, ""],
                package={key: _new_text},
            )
        except KeyError:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while updating {self._tag} data "
                f"for record ID {self._record_id} in the view.  Key {key} does not "
                f"exist in attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

        return {key: _new_text}

    def __do_read_text(
        self, entry: RAMSTKEntry, key: str, datatype: str
    ) -> Dict[str, Any]:
        """Read the text in a RAMSTKEntry() or Gtk.TextBuffer().

        :param entry: the RAMSTKEntry() or Gtk.TextBuffer() to read.
        :param key: the key for the entry to be read.
        :return: {_key, _new_text}; a dict containing the attribute key and
            the new value (text) for that key.
        """
        _new_text: Any = ""

        try:
            if str(datatype) == "gfloat":
                _new_text = float(entry.do_get_text())
            elif str(datatype) == "gint":
                _new_text = int(entry.do_get_text())
            elif str(datatype) == "gchararray":
                _new_text = str(entry.do_get_text())
        except (KeyError, ValueError):
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while reading {self._tag} data for "
                f"record ID {self._record_id} in the view.  Key {key} does not exist "
                f"in attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

        return {key: _new_text}


class RAMSTKPlotPanel(RAMSTKPanel):
    """The RAMSTKPlotPanel class.

    The attributes of a RAMSTKPlotPanel are:

    :ivar pltPlot: a RAMSTPlot() for the panels that embed a plot.
    :ivar lst_axis_labels: a list containing the labels for the plot's axes.  Index 0 is
        the abscissa (x-axis) and index 1 is the ordinate (y-axis).
    :ivar lst_legend: a list containing the text to display in the plot's legend.
    :ivar plot_title: the title of the plot.
    """

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.

    # Define public dict class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKPlotPanel.

        :return: None
        :rtype: None
        """
        super().__init__()

        # Initialize widgets.
        self.pltPlot: RAMSTKPlot = RAMSTKPlot()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.

        # Initialize public list instance attributes.
        self.lst_axis_labels: List[str] = [_("abscissa"), _("ordinate")]
        self.lst_legend: List[str] = []

        # Initialize public scalar instance attributes.
        self.plot_title: str = ""

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_clear_panel, "request_clear_views")

    def do_clear_panel(self) -> None:
        """Clear the contents of the RAMSTKPlot on a plot type panel.

        :return: None
        :rtype: None
        """
        self.pltPlot.axis.cla()
        self.pltPlot.figure.clf()
        self.pltPlot.plot.draw()

    def do_load_panel(self) -> None:
        """Load data into the RAMSTKPlot on a plot type panel.

        :return: None
        :rtype: None
        """
        self.pltPlot.do_make_title(self.plot_title)
        self.pltPlot.do_make_labels(
            self.lst_axis_labels[1], x_pos=-0.5, y_pos=0, set_x=False
        )

        self.pltPlot.do_make_legend(self.lst_legend)
        self.pltPlot.figure.canvas.draw()

    def do_make_panel(self) -> None:
        """Create a panel with a RAMSTKPlot().

        :return: None
        :rtype: None
        """
        _scrollwindow: Gtk.ScrolledWindow = Gtk.ScrolledWindow()
        _scrollwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        _scrollwindow.add(self.pltPlot.canvas)

        self.add(_scrollwindow)

    def do_set_callbacks(self) -> None:
        """Set the callback methods for RAMSTKTreeView().

        :return: None
        """

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set properties of the RAMSTKPanel() widgets.

        :return: None
        """
        super().do_set_properties(**{"bold": True, "title": self._title})


class RAMSTKTreePanel(RAMSTKPanel):
    """The RAMSTKTreePanel class.

    The attributes of a RAMSTKTreePanel are:

    :ivar tvwTreeView: a RAMSTKTreeView() for the panels that embed a treeview.
    :ivar _filtered_tree: boolean indicating whether or not to display the filtered
        RAMSTKTreeView() or the full RAMSTKTreeView().
    """

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.

    # Define public dict class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKTreePanel.

        :return: None
        :rtype: None
        """
        super().__init__()

        # Initialize widgets.
        self.tvwTreeView: RAMSTKTreeView = RAMSTKTreeView()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._filtered_tree: bool = False

        # Initialize public dict instance attributes.

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_clear_panel, "request_clear_views")
        pub.subscribe(self.do_load_panel, f"succeed_insert_{self._tag}")
        pub.subscribe(self.do_refresh_tree, f"lvw_editing_{self._tag}")
        pub.subscribe(self.do_refresh_tree, f"mvw_editing_{self._tag}")
        pub.subscribe(self.do_refresh_tree, f"wvw_editing_{self._tag}")
        pub.subscribe(self.on_delete_treerow, f"succeed_delete_{self._tag}")
        if self._select_msg is not None:
            pub.subscribe(self.do_load_panel, self._select_msg)

    def do_clear_panel(self) -> None:
        """Clear the contents of the RAMSTKTreeView on a tree type panel.

        :return: None
        :rtype: None
        """
        _model = self.tvwTreeView.get_model()
        try:
            _model.clear()
        except AttributeError:
            pass

    def do_load_panel(self, tree: treelib.Tree) -> None:
        """Load data into the RAMSTKTreeView on a tree type panel.

        :param tree: the treelib Tree containing the module to load.
        :return: None
        """
        try:
            self.tvwTreeView.unfilt_model.clear()
        except AttributeError:
            pass

        try:
            self.tvwTreeView.do_load_tree(tree, None)
            if self._filtered_tree:
                self.tvwTreeView.filt_model = self.tvwTreeView.unfilt_model.filter_new()
                self.tvwTreeView.filt_model.set_visible_func(self.filter_tree)
                self.tvwTreeView.set_model(self.tvwTreeView.filt_model)

            self.tvwTreeView.expand_all()
            _row = self.tvwTreeView.unfilt_model.get_iter_first()
            if _row is not None:
                self.tvwTreeView.selection.select_iter(_row)
                self.show_all()
        except TypeError:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while loading {self._tag} data for "
                f"Record ID {self._record_id} into the view.  One or more values from "
                f"the database was the wrong type for the column it was trying to load."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
        except ValueError:
            _method_name = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while loading {self._tag} data "
                f"for Record ID {self._record_id} into the view.  One or more values "
                f"from the database was missing."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

        pub.sendMessage("request_set_cursor_active")

    def do_load_treerow(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a row into the RAMSTKTreeView().

        :param node: the treelib Node() with the data to load.
        :param row: the parent row of the row to load.
        :return: _new_row; the row that was just populated with data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None
        _data: List[Any] = []

        try:
            # pylint: disable=unused-variable
            [[__, _entity]] = node.data.items()
            _attributes = _entity.get_attributes()
            for _key, _pos in self.tvwTreeView.position.items():
                _data.insert(_pos, _attributes[_key])

            _new_row = self.tvwTreeView.unfilt_model.append(row, _data)
        except (AttributeError, TypeError, ValueError) as _error:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = (
                f"{_method_name}: An error occurred when loading "
                f"{self._tag} {node.identifier}.  This might indicate it was missing "
                f"it's data package, some of the data in the package was missing, or "
                f"some of the data was the wrong type.  Row data was: {_data}.  Error "
                f"was: {_error}."
            )
            pub.sendMessage(
                "do_log_warning_msg",
                logger_name="WARNING",
                message=_error_msg,
            )
            _new_row = None

        return _new_row

    def do_make_panel(self) -> None:
        """Create a panel with a RAMSTKTreeView().

        :return: None
        """
        self.tvwTreeView.widgets = {
            _key: _value[1] for _key, _value in self.dic_attribute_widget_map.items()
        }
        self.tvwTreeView.editable = {
            _key: _value[6]["editable"]
            for _key, _value in self.dic_attribute_widget_map.items()
        }
        self.tvwTreeView.visible = {
            _key: _value[6]["visible"]
            for _key, _value in self.dic_attribute_widget_map.items()
        }
        self.tvwTreeView.datatypes = {
            _key: _value[8] for _key, _value in self.dic_attribute_widget_map.items()
        }

        _scrollwindow: Gtk.ScrolledWindow = Gtk.ScrolledWindow()
        _scrollwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        _scrollwindow.add(self.tvwTreeView)

        self.add(_scrollwindow)

    def do_make_treeview(self, **kwargs: Dict[str, Any]) -> None:
        """Make the RAMSTKTreeView() instance for this panel.

        :return: None
        :rtype: None
        """
        _bg_color: str = kwargs.get("bg_color", "#FFFFFF")  # type: ignore
        _fg_color: str = kwargs.get("fg_color", "#000000")  # type: ignore
        _fmt_file: str = kwargs.get("fmt_file", "")  # type: ignore
        _attrs = kwargs.get("attrs", {})

        self.tvwTreeView.do_parse_format(_fmt_file)
        self.tvwTreeView.do_make_model()

        for _key, _value in self.dic_attribute_widget_map.items():
            self.tvwTreeView.widgets[_key] = _value[1]
            self.tvwTreeView.cellprops[_key] = _attrs[_key][6]
            self.tvwTreeView.cellprops[_key]["bg_color"] = _bg_color
            self.tvwTreeView.cellprops[_key]["fg_color"] = _fg_color
            try:
                self.tvwTreeView.headings[_key] = _value[7]
            except IndexError:
                pass

        self.tvwTreeView.do_make_columns()

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def do_refresh_tree(self, node_id: List, package: Dict[str, Any]) -> None:
        """Update the module view RAMSTKTreeView() with attribute changes.

        This method is used to update a RAMSTKPanel() containing a
        RAMSTKTreeView() [generally the module view] whenever a work view
        widget is edited.  It is used to keep the data displayed in-sync.

        A dict 'package' is sent when a workview widget is edited/changed.

            `package` key: `package` value

        corresponds to:

            database field name: database field new value

        The key in the 'package' is used to find the value in
        _dic_attribute_updater corresponding to the data being changed.
        Position 2 of the _dic_attribute_updater value list is the nominal
        position in the RAMSTKTreeView() containing the same attribute data
        as the one being changed.

        :param node_id: unused in this method.
        :param package: the key:value for the data being updated.
        :return: None
        """
        _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore

        [[_key, _value]] = package.items()

        try:
            _position = self.tvwTreeView.position[_key]

            _model, _row = self.tvwTreeView.get_selection().get_selected()
            _model.set_value(_row, _position, _value)
        except KeyError:
            _error_msg = _(
                f"{_method_name}: An error occurred while refreshing {self._tag} data "
                f"for Record ID {self._record_id} in the view.  Key {_key} does not "
                f"exist in attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
        except TypeError:
            _error_msg = _(
                f"{_method_name}: An error occurred while refreshing {self._tag} data "
                f"for Record ID {self._record_id} in the view.  Data {_value} for "
                f"{_key} is the wrong type."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

    def do_set_callbacks(self) -> None:
        """Set the callback methods for RAMSTKTreeView().

        :return: None
        """
        self.tvwTreeView.dic_handler_id["changed"] = self.tvwTreeView.selection.connect(
            "changed", self._on_row_change
        )

        for _key, _value in self.dic_attribute_widget_map.items():
            try:
                if _value[3] is not None:
                    _value[1].connect(_value[2], _value[3], _key, _value[4])
            except KeyError:
                print(self._tag, type(_value))

    def do_set_cell_callbacks(self, message: str, columns: List[str]) -> None:
        """Set the callback methods for RAMSTKTreeView() cells.

        :param message: the PyPubSub message to broadcast on a successful edit.
        :param columns: the list of column numbers whose cells should have a callback
            function assigned.
        :return: None
        """
        for _key in columns:
            _cell = self.tvwTreeView.get_column(
                self.tvwTreeView.position[_key]
            ).get_cells()
            try:
                _cell[0].connect(
                    "edited",
                    self.on_cell_edit,
                    self.tvwTreeView.position[_key],
                    message,
                )
            except TypeError:
                _cell[0].connect(
                    "toggled",
                    self.on_cell_toggled,
                    self.tvwTreeView.position[_key],
                    message,
                )

    def do_set_headings(self) -> None:
        """Set the treeview headings depending on the selected row.

        It's used when the tree displays an aggregation of models such as
        the FMEA or PoF.  This method applies the appropriate headings when
        a row is selected.

        :return: None
        :rtype: None
        """
        _columns = self.tvwTreeView.get_columns()
        i = 0
        for _key in self.tvwTreeView.headings:
            _label = RAMSTKLabel(
                "<span weight='bold'>" + self.tvwTreeView.headings[_key] + "</span>"
            )
            _label.do_set_properties(
                height=-1, justify=Gtk.Justification.CENTER, wrap=True
            )
            _label.show_all()
            _columns[i].set_widget(_label)
            _columns[i].set_visible(self.tvwTreeView.visible[_key])

            i += 1

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set properties of the RAMSTKPanel() widgets.

        :return: None
        """
        super().do_set_properties(**{"bold": True, "title": self._title})

        self.tvwTreeView.set_enable_tree_lines(True)
        self.tvwTreeView.set_grid_lines(Gtk.TreeViewGridLines.BOTH)
        self.tvwTreeView.set_level_indentation(2)
        self.tvwTreeView.set_rubber_banding(True)

    def on_cell_change(
        self,
        cell: Gtk.CellRenderer,
        path: str,
        new_text: Gtk.TreeIter,
        key: str,
        message: str,
    ) -> None:
        """Handle edits of a Gtk.CellRendererCombo() in the panel's RAMSTKTreeview().

        :param cell: the Gtk.CellRenderer() that was edited.
        :param path: the RAMSTKTreeView() path of the Gtk.CellRenderer()
            that was edited.
        :param new_text: the new Gtk.TreeIter() selected in the
            Gtk.CellRendererCombo().  This is relative to the cell renderer's model,
            not the RAMSTKTreeView() model.
        :param key: the column key of the edited Gtk.CellRenderer().
        :param message: the PyPubSub message to publish.
        :return: None
        """
        try:
            _new_text = self.tvwTreeView.do_change_cell(
                cell,
                path,
                new_text,
                self.tvwTreeView.position[key],
            )

            pub.sendMessage(
                message,
                node_id=self._record_id,
                package={key: _new_text},
            )
        except KeyError:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while editing {self._tag} data "
                f"for record ID {self._record_id} in the view.  One or more keys could "
                f"not be found in the attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

    def on_cell_edit(
        self,
        cell: Gtk.CellRenderer,
        path: str,
        new_text: str,
        key: str,
        message: str,
    ) -> None:
        """Handle edits of Gtk.CellRendererText() in the panel's RAMSTKTreeview().

        :param cell: the Gtk.CellRenderer() that was edited.
        :param path: the RAMSTKTreeView() path of the Gtk.CellRenderer()
            that was edited.
        :param new_text: the new text in the edited Gtk.CellRenderer().
        :param key: the column key of the edited Gtk.CellRenderer().
        :param message: the PyPubSub message to publish.
        :return: None
        """
        try:
            self.tvwTreeView.do_edit_cell(
                cell,
                path,
                new_text,
                self.tvwTreeView.position[key],
            )

            pub.sendMessage(
                message,
                node_id=self._record_id,
                package={key: new_text},
            )
        except KeyError:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while editing {self._tag} data "
                f"for record ID {self._record_id} in the view.  One or more keys could "
                f"not be found in the attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

    # pylint: disable=unused-argument
    def on_cell_toggled(
        self, cell: Gtk.CellRenderer, path: str, key: str, message: str
    ) -> None:
        """Handle edits of the FMEA Work View RAMSTKTreeview() toggle cells.

        :param cell: the Gtk.CellRenderer() that was toggled.
        :param path: the RAMSTKTreeView() path of the Gtk.CellRenderer()
            that was toggled.
        :param key: the column key of the edited Gtk.CellRenderer().
        :param message: the PyPubSub message to publish.
        :return: None
        :rtype: None
        """
        _new_text = boolean_to_integer(cell.get_active())

        try:
            if not self.tvwTreeView.do_edit_cell(
                cell, path, _new_text, self.tvwTreeView.position[key]
            ):
                pub.sendMessage(
                    message, node_id=self._record_id, package={key: _new_text}
                )
        except KeyError:
            _method_name: str = inspect.currentframe().f_code.co_name  # type: ignore
            _error_msg = _(
                f"{_method_name}: An error occurred while editing {self._tag} data "
                f"for record ID {self._record_id} in the view.  One or more keys could "
                f"not be found in the attribute dictionary."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def on_delete_treerow(self, tree: treelib.Tree) -> None:
        """Update the RAMSTKTreeView after deleting a line item.

        :param tree: the treelib Tree() containing the workflow module data.
        :return: None
        """
        _model, _row = self.tvwTreeView.selection.get_selected()

        if self._filtered_tree:
            _model = self.tvwTreeView.filt_model.get_model()
            _row = self.tvwTreeView.filt_model.convert_iter_to_child_iter(_row)
            self.tvwTreeView.unfilt_model.remove(_row)
            self.tvwTreeView.filt_model.refilter()
        else:
            self.tvwTreeView.unfilt_model.remove(_row)
            _row = _model.get_iter_first()
            if _row is not None:
                self.tvwTreeView.selection.select_iter(_row)
                self.show_all()

        pub.sendMessage("request_set_cursor_active")

    def on_insert(self, data: Any) -> None:
        """Add row to module view for newly added work stream element.

        :param data: the data package for the work stream element to add.
        :return: None
        """
        _model, _row = self.tvwTreeView.selection.get_selected()

        # When inserting a child record, the selected row becomes the parent
        # row.
        if self._record_id == self._parent_id:
            _prow = _row
        # When inserting a sibling record, use the parent of the selected
        # row.
        else:
            _prow = _model.iter_parent(_row)

        self.tvwTreeView.do_insert_row(data, _prow)

        pub.sendMessage("request_set_cursor_active")

    def on_row_change(self, selection: Gtk.TreeSelection) -> Dict[str, Any]:
        """Get the attributes for the newly selected row.

        :param selection: the Gtk.TreeSelection() for the new row.
        :return: _attributes; the dict of attributes and value for the item
            in the selected row.  The key is the attribute name, the value is
            the attribute value.  Pulling them from the RAMSTKTreeView()
            ensures uncommitted changes are always selected.
        """
        selection.handler_block(self.tvwTreeView.dic_handler_id["changed"])

        _attributes: Dict[str, Any] = {}

        _model, _row = selection.get_selected()
        if _row is not None:
            for (
                _key,
                __,  # pylint: disable=unused-variable
            ) in self.dic_attribute_widget_map.items():
                _attributes[_key] = _model.get_value(
                    _row,
                    self.tvwTreeView.position[_key],
                )

        selection.handler_unblock(self.tvwTreeView.dic_handler_id["changed"])

        return _attributes

    def _do_set_attributes(self, node_id, package) -> None:
        """Set the attributes of the record associated with node ID.

        This is a helper method to use with database view models.  Since database
        view models are comprised of an aggregate of database tables, the selected
        row in the RAMSTKTreeView will determine which table needs updating.  This
        method susses that out and sends the correct message to cause the correct
        table to be updated.

        :param node_id: the ID of the record in the RAMSTK Program database table whose
            attributes are to be set.
        :param package: the key:value pair of the attribute to set.
        :return: None
        :rtype: None
        """
        pub.sendMessage(
            f"lvw_editing_{self._tag}",
            node_id=node_id,
            package=package,
        )



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.validation.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Validation Record Model."""

# Standard Library Imports
from datetime import date, timedelta

# Third Party Imports
# noinspection PyPackageRequirements
from sqlalchemy import Column, Date, Float, ForeignKey, Integer, String

# RAMSTK Package Imports
from ramstk.analyses.statistics import do_calculate_beta_bounds
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKValidationRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent table ramstk_validation in RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_revision.
    """

    __defaults__ = {
        "acceptable_maximum": 0.0,
        "acceptable_mean": 0.0,
        "acceptable_minimum": 0.0,
        "acceptable_variance": 0.0,
        "confidence": 95.0,
        "cost_average": 0.0,
        "cost_ll": 0.0,
        "cost_maximum": 0.0,
        "cost_mean": 0.0,
        "cost_minimum": 0.0,
        "cost_ul": 0.0,
        "cost_variance": 0.0,
        "date_end": date.today() + timedelta(days=30),
        "date_start": date.today(),
        "description": "",
        "measurement_unit": "",
        "name": "",
        "status": 0.0,
        "task_type": "",
        "task_specification": "",
        "time_average": 0.0,
        "time_ll": 0.0,
        "time_maximum": 0.0,
        "time_mean": 0.0,
        "time_minimum": 0.0,
        "time_ul": 0.0,
        "time_variance": 0.0,
    }
    __tablename__ = "ramstk_validation"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    validation_id = Column(
        "fld_validation_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    acceptable_maximum = Column(
        "fld_acceptable_maximum", Float, default=__defaults__["acceptable_maximum"]
    )
    acceptable_mean = Column(
        "fld_acceptable_mean", Float, default=__defaults__["acceptable_mean"]
    )
    acceptable_minimum = Column(
        "fld_acceptable_minimum", Float, default=__defaults__["acceptable_minimum"]
    )
    acceptable_variance = Column(
        "fld_acceptable_variance", Float, default=__defaults__["acceptable_variance"]
    )
    confidence = Column("fld_confidence", Float, default=__defaults__["confidence"])
    cost_average = Column(
        "fld_cost_average", Float, default=__defaults__["cost_average"]
    )
    cost_ll = Column("fld_cost_ll", Float, default=__defaults__["cost_ll"])
    cost_maximum = Column(
        "fld_cost_maximum", Float, default=__defaults__["cost_maximum"]
    )
    cost_mean = Column("fld_cost_mean", Float, default=__defaults__["cost_mean"])
    cost_minimum = Column(
        "fld_cost_minimum", Float, default=__defaults__["cost_minimum"]
    )
    cost_ul = Column("fld_cost_ul", Float, default=__defaults__["cost_ul"])
    cost_variance = Column(
        "fld_cost_variance", Float, default=__defaults__["cost_variance"]
    )
    date_end = Column("fld_date_end", Date, default=__defaults__["date_end"])
    date_start = Column("fld_date_start", Date, default=__defaults__["date_start"])
    description = Column("fld_description", String, default=__defaults__["description"])
    measurement_unit = Column(
        "fld_measurement_unit", String(256), default=__defaults__["measurement_unit"]
    )
    name = Column("fld_name", String(256), default=__defaults__["name"])
    status = Column("fld_status", Float, default=__defaults__["status"])
    task_specification = Column(
        "fld_task_specification",
        String(512),
        default=__defaults__["task_specification"],
    )
    task_type = Column("fld_type", String(256), default=__defaults__["task_type"])
    time_average = Column(
        "fld_time_average", Float, default=__defaults__["time_average"]
    )
    time_ll = Column("fld_time_ll", Float, default=__defaults__["time_ll"])
    time_maximum = Column(
        "fld_time_maximum", Float, default=__defaults__["time_maximum"]
    )
    time_mean = Column("fld_time_mean", Float, default=__defaults__["time_mean"])
    time_minimum = Column(
        "fld_time_minimum", Float, default=__defaults__["time_minimum"]
    )
    time_ul = Column("fld_time_ul", Float, default=__defaults__["time_ul"])
    time_variance = Column(
        "fld_time_variance", Float, default=__defaults__["time_variance"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.

    def get_attributes(self):
        """Retrieve current values of RAMSTKValidation data model attributes.

        :return: {revision_id, validation_id, acceptable_maximum,
                  acceptable_mean, acceptable_minimum, acceptable_variance,
                  confidence, cost_average, cost_ll. cost_maximum, cost_mean,
                  cost_minimum, cost_l, cost_variance, date_end, date_start,
                  description, measurement_unit_id, status_id, task_type_id,
                  task_specification, time_average, time_ll, time_maximum,
                  time_mean, time_minimum, time_ul, time_variance} pairs.
        :rtype: dict
        """
        _attributes = {
            "revision_id": self.revision_id,
            "validation_id": self.validation_id,
            "acceptable_maximum": self.acceptable_maximum,
            "acceptable_mean": self.acceptable_mean,
            "acceptable_minimum": self.acceptable_minimum,
            "acceptable_variance": self.acceptable_variance,
            "confidence": self.confidence,
            "cost_average": self.cost_average,
            "cost_ll": self.cost_ll,
            "cost_maximum": self.cost_maximum,
            "cost_mean": self.cost_mean,
            "cost_minimum": self.cost_minimum,
            "cost_ul": self.cost_ul,
            "cost_variance": self.cost_variance,
            "date_end": self.date_end,
            "date_start": self.date_start,
            "description": self.description,
            "measurement_unit": self.measurement_unit,
            "name": self.name,
            "status": self.status,
            "task_specification": self.task_specification,
            "task_type": self.task_type,
            "time_average": self.time_average,
            "time_ll": self.time_ll,
            "time_maximum": self.time_maximum,
            "time_mean": self.time_mean,
            "time_minimum": self.time_minimum,
            "time_ul": self.time_ul,
            "time_variance": self.time_variance,
        }

        return _attributes

    def calculate_task_time(self):
        """Calculate the mean, standard error, and bounds on the task time.

        These values are calculated assuming a beta distribution (typical
        project management assumption).

        :return: None
        :rtype: None
        """
        (self.time_ll, self.time_mean, self.time_ul, _sd) = do_calculate_beta_bounds(
            self.time_minimum, self.time_average, self.time_maximum, self.confidence
        )

        self.time_variance = _sd ** 2.0

    def calculate_task_cost(self):
        """Calculate the mean, standard error, and bounds on the task cost.

        These values are calculated assuming a beta distribution (typical
        project management assumption).

        :return: None
        :rtype: None
        """
        (self.cost_ll, self.cost_mean, self.cost_ul, _sd) = do_calculate_beta_bounds(
            self.cost_minimum, self.cost_average, self.cost_maximum, self.confidence
        )

        self.cost_variance = _sd ** 2.0



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKMilHdbk217FRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    hardware_id: Any
    A1: Any
    A2: Any
    B1: Any
    B2: Any
    C1: Any
    C2: Any
    lambdaBD: Any
    lambdaBP: Any
    lambdaCYC: Any
    lambdaEOS: Any
    piA: Any
    piC: Any
    piCD: Any
    piCF: Any
    piCR: Any
    piCV: Any
    piCYC: Any
    piE: Any
    piF: Any
    piI: Any
    piK: Any
    piL: Any
    piM: Any
    piMFG: Any
    piN: Any
    piNR: Any
    piP: Any
    piPT: Any
    piQ: Any
    piR: Any
    piS: Any
    piT: Any
    piTAPS: Any
    piU: Any
    piV: Any
    def get_attributes(self): ...
